/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_acl_defs.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief API for using the ACL layer.
 * 
 * Contains:
 * - Initialization & Definition of maximal ACL packet payload sizes
 * - API for sending ACL data
 * - API for accessing ACL packet header data
 */
#ifndef _BT_ACL_DEFS_H_
#define _BT_ACL_DEFS_H_

#include <sys/types.h>
#include <bt/bt_defs.h>
#include <bt/bt_hci_api.h>

/**
 * \name ACL Payload Sizes
 * 
 * These are the defines of the different ACL payload sizes
 */
//@{

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM1.
 */
#define BT_ACL_PAYLOAD_SIZE_DM1			16

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH1.
 */
#define BT_ACL_PAYLOAD_SIZE_DH1			26

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM3.
 */
#define BT_ACL_PAYLOAD_SIZE_DM3			121

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH3.
 */
#define BT_ACL_PAYLOAD_SIZE_DH3			183

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM5.
 */
#define BT_ACL_PAYLOAD_SIZE_DM5			224

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH5.
 */
#define BT_ACL_PAYLOAD_SIZE_DH5			339

/**
 * Size of the HCI packet header that gets attached to each acl packet
 * 
 * Note: includes the hci-type field.
 */
#define BT_ACL_HCI_PKT_HDR_LEN		5
//@}


/**
 * \name Initialization
 */
//@{

/** \brief Initializes the ACL layer.
 * 
 * Sets the ACL packet type(s) to use and creates the first ACL buffer.
 * 
 * Note: When dealing with ACL data packets, this function should be called
 * immediately after the bt-stack has been initialized! The reason for this
 * is that setting the ACL packet types to use determines the size of
 * the ACL buffers to be used on the receiving side.
 * 
 * Note: ACL packet type definition can be found in bt_hci_defs.h
 * 
 * \param stack The bt-stack to use
 * \param pkt_types ACL packet type(s) to use
 * 
 * \returns Maximal ACL pkt payload size, corresponding to the packet types
 * chosen.
 * 
 * \see bt_hci_defs.h
 */
u_short bt_acl_init(struct btstack* stack, u_short pkt_types);
//@}


/**
 * \name Sending ACL Data
 * 
 * The following functions define the API provided to send ACL data.
 */
//@{

/**
 * \brief Sends an ACL data packet.
 * 
 * \param stack Pointer to the bt-stack to use
 * \param con_hdl Connection handle of the desired baseband connection
 * \param pb_flag Packet Boundary Flag #HCI_PB_FIRST_FRAGMENT or #HCI_PB_CONTD_FRAGMENT
 * \param bc_flag Broadcast Flag #HCI_BC_UNICAST, #HCI_BC_ACTIVE or #HCI_BC_PICONET
 * \param length Amount of ACL data to send
 * 
 * \returns 0 or possible errors:
 * - #BT_ERR_NO_CON
 */
long bt_acl_send_packet(struct btstack* stack,
						bt_hci_con_handle_t con_hdl,
						u_char pb_flag,
						u_char bc_flag,
						u_char* data,
						u_short len);

/**
 * \brief Acquires a given baseband connection to send ACL data.
 * 
 * Basically, this function provides unique access to the bt uart and
 * writes the information needed by the hci layer to send acl data over
 * a given baseband communication link to the bt controller.
 * 
 * \param stack The bt-stack to use
 * \param con_handle Connection handle of the desired baseband connection
 * \param pb_flag Packet Boundary Flag #HCI_PB_FIRST_FRAGMENT or #HCI_PB_CONTD_FRAGMENT
 * \param bc_flag Broadcast Flag #HCI_BC_UNICAST, #HCI_BC_ACTIVE or #HCI_BC_PICONET
 * \param length Amount of ACL data to send
 * 
 * \returns 0 or possible errors:
 * - #BT_ERR_NO_CON
 * 
 * Note: ACL data can be sent to the con handle #BT_HCI_CON_HANDLE_OWN, i.e.
 * to the sending device itself.
 * 
 * \see
 * #bt_acl_lowlevel_write \n
 * #bt_acl_write_lock_release
 */
long bt_acl_acquire_write_lock(struct btstack* stack,
						bt_hci_con_handle_t con_handle,
						u_char pb_flag,
						u_char bc_flag,
						u_short length);

/**
 * \brief Writes ACL data to the previousely acquired baseband connection.
 * 
 * Use #bt_acl_acquire_write_lock to get access to a desired baseband connection.
 * Write your data, and release the connection using #bt_acl_write_lock_release.
 * 
 * \param stack Pointer to the bt-stack in use
 * \param data Pointer to the data to be written
 * \param len Number of bytes to be written
 * 
 * \see
 * #bt_acl_acquire_write_lock \n
 * #bt_acl_write_lock_release
 */
inline long bt_acl_lowlevel_write(struct btstack* stack, u_char* data, u_short len);

/**
 * \brief Releases a given baseband connection.
 * 
 * Call this function after you've written your data to the bt-controller
 * using #bt_acl_lowlevel_write.
 * 
 * \param stack Pointer to the bt-stack in use
 * 
 * \see
 * #bt_acl_acquire_write_lock \n
 * #bt_acl_lowlevel_write
 */
inline void bt_acl_write_lock_release(struct btstack* stack);
//@}


/**
 * \name ACL Packet Accessors
 * 
 * \brief API functions for accessing the header fileds of received
 * ACL packets.
 */
//@{

/**
 * \brief Returns a pointer to the payload of an ACL packet
 * 
 * \param pkt Pointer to the received ACL packet
 */
inline u_char* bt_acl_get_data_pointer(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the con handle the ACL packet passed was received on.
 * 
 * \param pkt Pointer to the received ACL packet
 */
inline bt_hci_con_handle_t bt_acl_get_con_handle(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the fragmentation flag of the ACL packet passed.
 * 
 * \param pkt Pointer to the received ACL packet
 */
inline u_char bt_acl_get_pb_flag(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the broadcast flag of the ACL packet passed.
 * 
 * \param pkt Pointer to the received ACL packet
 */
inline u_char bt_acl_get_bc_flag(struct bt_hci_pkt_acl* pkt);

//@}


/**
 * \brief Returns current setting of ACL packet types to use.
 * 
 * Note: The ACL packet types to use has to be defined during
 * initialization.
 * 
 * \param stack Pointer to the bt-stack in use
 * 
 * \returns packet types in use (see 
 * 
 * \see
 * #bt_acl_init \n
 * bt_hci_defs.h
 */
inline u_short bt_acl_get_pkt_types(struct btstack* stack);

/**
 * \brief Returns the maximal ACL payload size.
 * 
 * Note: Maximal ACL payload size depends on the ACL packet types to
 * use. This has to be defined during initialization.
 * 
 * \param stack Pointer to the bt-stack in use
 * 
 * \returns maximal ACL payload size
 * 
 * \see
 * #bt_acl_init \n
 * bt_hci_defs.h
 */
inline u_short bt_acl_get_max_pkt_len(struct btstack* stack);

#endif //_BT_ACL_DEFS_H_
