/*
 * Copyright (C) 2000-2004 by ETH Zurich
 * Copyright (C) 2004 by Ole Reinhardt <ole.reinhardt@embedded-it.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 * 
 */
 
 
/**
 * @file bt_rfcomm.h
 *\author Ole Reinhardt <ole.reinhardt@embedded-it.de>
 *\date 25.01.2005
 * 
 * @brief Implementation of the RFCOMM protocol for BTNut. 
 * A subset of the ETSI TS 07.10 standard with some Bluetooth-specific 
 * adaptations.
 * 
 * See Bluetooth 1.1 specification for better understanding of l2cap.
 *
 * Some parts are taken from or inspired by the lwBT Bluetooth stack.
 * 
 * (lwBT author: Conny Ohult <conny@sm.luth.se>)
 * Copyright (c) 2003 EISLAB, Lulea University of Technology.)
 */

 
#ifndef __BT_RFCOMM_H__
#define __BT_RFCOMM_H__

#include <bt/bt_hci_defs.h>

#define BT_RFCOMM_MAX_CHANNELS 30       // We will always reserv the same amount of active and listening channels

#define BT_RFCOMM_CHNR_INVALID 0xFF
#define BT_RFCOMM_DLCI_INVALID 0xFF
#define BT_RFCOMM_DLCI_MUX   0x00

// Control field values      bit no.       1 2 3 4 5   6 7 8
#define BT_RFCOMM_SABM       0x3F       // 1 1 1 1 P/F 1 0 0
#define BT_RFCOMM_UA         0x73       // 1 1 0 0 P/F 1 1 0
#define BT_RFCOMM_DM         0x0F       // 1 1 1 1 P/F 0 0 0
#define BT_RFCOMM_DM_PF      0x1F
#define BT_RFCOMM_DISC       0x53       // 1 1 0 0 P/F 0 1 1
#define BT_RFCOMM_UIH        0xEF       // 1 1 1 1 P/F 1 1 1
#define BT_RFCOMM_UIH_PF     0xFF

// Command or response marker

#define BT_RFCOMM_CMD        0x01
#define BT_RFCOMM_RSP        0x00

// Multiplexer message types 
#define BT_RFCOMM_PN_CMD     0x83
#define BT_RFCOMM_PN_RSP     0x81
#define BT_RFCOMM_TEST_CMD   0x23
#define BT_RFCOMM_TEST_RSP   0x21
#define BT_RFCOMM_FCON_CMD   0xA3
#define BT_RFCOMM_FCON_RSP   0xA1
#define BT_RFCOMM_FCOFF_CMD  0x63
#define BT_RFCOMM_FCOFF_RSP  0x61
#define BT_RFCOMM_MSC_CMD    0xE3
#define BT_RFCOMM_MSC_RSP    0xE1
#define BT_RFCOMM_RPN_CMD    0x93
#define BT_RFCOMM_RPN_RSP    0x91
#define BT_RFCOMM_RLS_CMD    0x53
#define BT_RFCOMM_RLS_RSP    0x51
#define BT_RFCOMM_NSC_RSP    0x11

// Length of rfcomm hdr with 1 or 2 byte lengh field excluding credit 
#define BT_RFCOMM_HDR_LEN_1  3
#define BT_RFCOMM_HDR_LEN_2  4

#define BT_RFCOMM_FCS_LEN    1

// Length of a multiplexer message 
#define BT_RFCOMM_MSGHDR_LEN   2
#define BT_RFCOMM_PNMSG_LEN    8
#define BT_RFCOMM_MSCMSG_LEN   2
#define BT_RFCOMM_RPNMSG_LEN   8
#define BT_RFCOMM_NSCMSG_LEN   1
#define BT_RFCOMM_FCONMSG_LEN  0
#define BT_RFCOMM_FCOFFMSG_LEN 0
#define BT_RFCOMM_RLSMSG_LEN   2
#define BT_RFCOMM_EMPTYMSG_LEN 0

// Length of a frame 
#define BT_RFCOMM_DM_LEN      4
#define BT_RFCOMM_SABM_LEN    4
#define BT_RFCOMM_DISC_LEN    4
#define BT_RFCOMM_UA_LEN      4
#define BT_RFCOMM_UIH_LEN     3
#define BT_RFCOMM_UIHCRED_LEN 4

// Default convergence layer 
#define BT_RFCOMM_DEF_CL      0x0F      // Credit based flow control
#define BT_RFCOMM_DEF_MFS     (BT_L2CAP_MTU_DEFAULT - BT_RFCOMM_HDR_LEN_2 - BT_RFCOMM_FCS_LEN)        //1019      //669  gives MTU 674// TODO: What is a good value? (PPP MRU * 2) + 8
#define BT_RFCOMM_DEF_MTU     (u_short)(BT_RFCOMM_DEF_MFS + BT_RFCOMM_HDR_LEN_2 + BT_RFCOMM_FCS_LEN)
#define BT_RFCOMM_DEF_CREDITS 0x07      // Default window size... (credits) // TODO Should this be 0 or 7?
#define BT_RFCOMM_DEF_PRIO    0x07      // Default priority. It's depending on the dlci value. 
                                        // See ETSI TS 07.10 standard subclause 5.6 for detailed information.
#define BT_RFCOMM_DEF_TO      20000     // Default timeout in miliseconds


// Default port settings for a communication link 
#define BT_RFCOMM_COM_BR_DEF        0x03 // Baud rate (9600 bit/s)
#define BT_RFCOMM_COM_CFG_DEF       0x03 // Data bits (8 bits), stop bits (1), parity (no parity) and parity type
#define BT_RFCOMM_COM_FC_DEF        0x00 // Flow control (no flow ctrl)
#define BT_RFCOMM_COM_XON_DEF       0x00 // No flow control (default DC1)
#define BT_RFCOMM_COM_XOFF_DEF      0x00 // No flow control (default DC3)

// FCS calc 
#define BT_RFCOMM_CODE_WORD         0xE0 // pol = x8+x2+x1+1
#define BT_RFCOMM_CRC_CHECK_LEN     3
#define BT_RFCOMM_UIHCRC_CHECK_LEN  2

// rfcomm configuration parameter masks
#define BT_RFCOMM_CFG_INITIATOR     0x01      // we are the Initiator
#define BT_RFCOMM_CFG_FC            0x02      // flow control bit
#define BT_RFCOMM_CFG_MSC_OUT       0x04      // modem status configured for output?
#define BT_RFCOMM_CFG_MSC_IN        0x08      // modem status configured for input?

// rfcomm remote port settings (baudrate)

#define BT_RFCOMM_RPN_BAUD_2400     0x00
#define BT_RFCOMM_RPN_BAUD_4800     0x01
#define BT_RFCOMM_RPN_BAUD_7200     0x02
#define BT_RFCOMM_RPN_BAUD_9600     0x03
#define BT_RFCOMM_RPN_BAUD_19200    0x04
#define BT_RFCOMM_RPN_BAUD_38400    0x05
#define BT_RFCOMM_RPN_BAUD_57600    0x06
#define BT_RFCOMM_RPN_BAUD_115200   0x07
#define BT_RFCOMM_RPN_BAUD_230400   0x08

// Remote line status errors

#define BT_RFCOMM_RLS_ERROR         0x01
#define BT_RFCOMM_RLS_ERR_OVERRUN   0x02
#define BT_RFCOMM_RLS_ERR_PARITY    0x04
#define BT_RFCOMM_RLS_ERR_FRAMING   0x08

// error states

#define BT_RFCOMM_ERR_SUCCESS       0x00
#define BT_RFCOMM_ERR_NO_RESSOURCE  0x01
#define BT_RFCOMM_ERR_TIMEOUT       0x02
#define BT_RFCOMM_ERR_INVALID_CH    0x03
#define BT_RFCOMM_ERR_OOLCR         0x04
#define BT_RFCOMM_ERR_CON_CLOSED    0x05
#define BT_RFCOMM_ERR_CON_REFUSED   0x06
#define BT_RFCOMM_ERR_DATASIZE      0x07
#define BT_RFCOMM_ERR_SESSION_EXISTS 0x08


#define BT_RFCOMM_ERR_OFFS_L2CAP    0x20


// type of frames

#define BT_RFCOMM_FRAME_UIH   0x00
#define BT_RFCOMM_FRAME_UI    0x01
#define BT_RFCOMM_FRAME_I     0x02


// signal definition for MSC command:

#define BT_RFCOMM_SIG_FC        0x02
#define BT_RFCOMM_SIG_RTC       0x04       
#define BT_RFCOMM_SIG_RTR       0x08
#define BT_RFCOMM_SIG_IC        0x40
#define BT_RFCOMM_SIG_DV        0x80

#define BT_RFCOMM_SIG_DEFAULT   (BT_RFCOMM_SIG_RTC | BT_RFCOMM_SIG_RTR | BT_RFCOMM_SIG_DV)

#define BT_RFCOMM_CH_CLOSED     0x01
#define BT_RFCOMM_CH_LISTEN     0x02
#define BT_RFCOMM_CH_OPEN       0x03
#define BT_RFCOMM_CH_CONNECTED  0x04

#define BT_RFCOMM_CH_W4_MULTIPLEXER 0x05
#define BT_RFCOMM_CH_W4_SABM_RSP    0x06
#define BT_RFCOMM_CH_CFG            0x07
#define BT_RFCOMM_CH_W4_DISC_RSP    0x08

// Connection callback types

#define BT_RFCOMM_CONNECT           0x01
#define BT_RFCOMM_DISCONNECT        0x02

// Callback types. They are registered with the open or listen routine.

#define BT_RFCOMM_RCV_CB void (*rcv_cb) (u_char dlci, u_char *payload, u_short len, void *arg)
#define BT_RFCOMM_CON_CB void (*con_cb) (u_char dlci, u_char type, void *arg)
#define BT_RFCOMM_LINE_CB void (*line_cb) (u_char dlci, u_char flags, void *arg)
#define BT_RFCOMM_CREDIT_CB void (*credit_cb) (u_char dlci, u_char credits, void *arg)

#define BT_RFCOMM_NO_TIMEOUT        0xFFFFFFFF

//  RFCOMM PACKET FORMAT

// 
// +---------+---------+---------+--------...------+---------+
// | address | control | length  |   information   |   FCS   |
// +---------+---------+---------+--------...------+---------+
//      1         1      1 (or2)        length          1      bytes
//

// The RFCOMM packet header
/*
struct bt_rfcomm_header {
  u_char  addr;
  u_char  ctrl;
  u_short len;
  u_char credit;
};
*/

// The RFCOMM protocol control block / channel information
typedef struct bt_rfcomm_channel {

    u_char state;               // RFCOMM channel state 

    u_char dlci;                // dlci of this channel ==> channel_nr >> 1 | direction bit

    u_char convergency_layer;   // Convergence layer 
    u_char priority;            // Connection priority 
    u_short max_framesize;      // Maximum frame size 
    u_char local_credits;       // Number of local credits 
    u_char remote_credits;      // Number of remote credits
    u_char config;              // See BT_RFCOMM_CFG_XXXXXX

    u_char uih_in_fcs;          // Frame check sequence for uih frames (P/F bit = 0)
    u_char uihpf_in_fcs;        // Frame check sequence for uih frames (P/F bit = 1)
    u_char uih_out_fcs;         // Frame check sequence for uih frames (P/F bit = 0)
    u_char uihpf_out_fcs;       // Frame check sequence for uih frames (P/F bit = 1)
    
    u_char response_type;       // Type of response (command)
    HANDLE sig_response;        // Signal that a response has been received.
    HANDLE sig_connected;       
    HANDLE sig_credits;         // signal will be set if we get credits
    
    BT_RFCOMM_RCV_CB;
    BT_RFCOMM_CON_CB;
    BT_RFCOMM_LINE_CB;          // line callback. Will be called on change of the status lines
    BT_RFCOMM_CREDIT_CB;        // will be called if credits became low
    void   *cb_arg;
    u_char credit_limit;
    struct bt_rfcomm_listen* listen;
} bt_rfcomm_channel_t;

typedef struct bt_rfcomm_listen {
    u_char channel_nr;
    BT_RFCOMM_RCV_CB;
    BT_RFCOMM_CON_CB;
    BT_RFCOMM_LINE_CB;          // line callback. Will be called on change of the status lines
    BT_RFCOMM_CREDIT_CB;        // will be called if credits became low
    u_char credit_limit;
    HANDLE sig_connected;
    void   *cb_arg;
} bt_rfcomm_listen_t;

typedef struct bt_rfcomm_l2cap_pkt_buf_entry {
    struct bt_l2cap_acl_pkt *pkt;
    u_short sequence_nr;
    u_char state;
} bt_rfcomm_l2cap_pkt_buf_entry_t;

typedef struct bt_rfcomm_stack {
    struct bt_l2cap_stack    *l2cap_stack;
    struct bt_rfcomm_channel *channels[BT_RFCOMM_MAX_CHANNELS * 2];
    struct bt_rfcomm_listen  *listening[BT_RFCOMM_MAX_CHANNELS+1];
    struct bt_l2cap_acl_pkt  *send_buffer;
    u_char max_channel;
    u_char max_dlci;
    u_char initiator;           // 0 if we are the responder, 1 of we are initiator

    u_char service_nr;
    u_short l2cap_channel_id;   // channel id for L2CAP
    void   *handle;
    u_short l2cap_sequence_nr;

    struct bt_rfcomm_l2cap_pkt_buf_entry *buffer;
    u_char buffer_size;

    HANDLE sig_l2cap_data_avail;
    HANDLE sig_l2cap_connected;
    u_char l2cap_disconnected;
} bt_rfcomm_stack_t;

// Macro definitions

#define BT_RFCOMM_DLCI2CH(dlci)         (dlci >> 1)

// Function definitions

u_char bt_rfcomm_send_credits(u_char dlci, u_char credits);
u_char bt_rfcomm_send(u_char dlci, u_char *data, u_short length);
u_char bt_rfcomm_test(u_long timeout);
u_char bt_rfcomm_start_session(bt_addr_t bt_addr, u_char page_scan_rep_mode, u_short clock_offset);
// bt_rfcomm_listen: Timeout = BT_RFCOMM_NO_TIMEOUT will return imediatly. 
u_char bt_rfcomm_listen(u_char channel_nr, BT_RFCOMM_CON_CB, BT_RFCOMM_RCV_CB, BT_RFCOMM_LINE_CB, BT_RFCOMM_CREDIT_CB, u_char credit_limit, void* cb_arg, u_long timeout);
u_char bt_rfcomm_connect(u_char channel_nr, BT_RFCOMM_CON_CB, BT_RFCOMM_RCV_CB, BT_RFCOMM_LINE_CB, BT_RFCOMM_CREDIT_CB, u_char credit_limit, void* cb_arg);
u_char bt_rfcomm_disconnect(u_char dlci);
struct bt_rfcomm_stack *bt_rfcomm_init(struct bt_l2cap_stack *l2cap_stack, u_short max_mfs, u_char nr_channels, u_char nr_l2cap_buffer);

#endif
