/*
 * Copyright (C) 2005 Mario Strasser <mast@gmx.net>,
 *                    Swiss Federal Institute of Technology (ETH) Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: bmac.h,v 1.2 2006/09/22 15:17:28 freckle Exp $
 */

/**
 * @file bmac.h
 * @author Mario Strasser <mast@gmx.net>
 * @brief B-MAC implementation for the CC1000 radio based on
 * [Polastre et al., Versatile Low Power Media Access for Wireless
 * Sensor Networks] and [Polastre, Sensor Network Media Access Design]
 */

#ifndef _BMAC_H_
#define _BMAC_H_

#include <sys/types.h>
#include <cc/mac.h>

/**
 * define this if debug output should be printed
 */
/* #define DEBUG_BMAC */

#ifdef DEBUG_BMAC
#include <stdio.h>
#define BMAC_DEBUG(str, ...) printf_P(PSTR(str), ## __VA_ARGS__)
#else
#define BMAC_DEBUG(str, ...) {}
#endif

extern u_long bmac_recv_count;
extern u_long bmac_failed_count;

/**
 * Initialises the B-MAC library.
 * 
 * @param address [in] mac address of the node
 * @return 0 on success, -1 otherwise.
 */
int bmac_init(u_short address);

/**
 * Sets the duration of a LPL interval. If the interval is set to zero,
 * LPL is disabled and the radio is permanently on.
 * 
 * @param interval [in] duration of the interval in ms
 */
void bmac_set_lpl_interval(u_short interval);

/**
 * Sets the maximal duration of the initial backoff.
 * 
 * @param max_duration [in] maximal initial backoff-time
 */
void bmac_set_initial_backoff(u_char max_duration);

/**
 * Sets the maximal duration of the congestion backoff.
 * 
 * @param max_duration [in] maximal congestion backoff-time
 */
void bmac_set_congestion_backoff(u_char max_duration);

/**
 * Enables/disables the automatic acknowledgment of unicast packets.
 * 
 * @param enable_ack [in] true (1) to enable or false (0) to disable
 *                   automatic acknowledgments
 */
void bmac_set_ack(u_char enable_ack);

/**
 * Queries the status pf the automatic acknowledgments of unicast packets.
 * 
 * @return true (1) if automatic acknowledgments are enabled, alse (0) 
 * if automatic acknowledgments are disabled
 *                   
 */
u_char bmac_get_ack(void);

/**
 * Enables/disables the deactivation of the timer interrupt while sending.
 * 
 * @param disable_timer [in] true (1) to enable or false (0) to disable
 *                      the deactivation mechanism
 */
void bmac_disable_timer_interrupt(int disable_timer);

/**
 * Enables/disables led feedback.
 * 
 * @param enable_led [in] true (1) to enable or false (0) to disable 
 *                   led feedback
 */
void bmac_enable_led(int enable_led);


/**
 * Enables/disables the usage of the SPI interrupt for sending.
 *
 * @param use_irq [in] true (1) to enable the usage of the SPI interrupt
 *                or false (0) to use busy-waiting
 */
void bmac_use_interrupt_to_send(int use_irq);

/**
 * Returns the configured mac address.
 * 
 * @return configured mac address.
 */
u_short bmac_get_address(void);

/**
 * Returns the maximal payload size of a B-MAC packet.
 * 
 * @return maximal payload size of a B-MAC packet.
 */
u_short bmac_get_mtu(void);

/**
 * Sends a B-MAC packet.
 * 
 * @param dst [in] destination address
 * @param data [in] payload of the packet
 * @param data_len [in] payload size
 * @return 0 on success, -1 otherwise.
 */
 
int bmac_send(u_short dst, u_char *data, u_short data_len);

/**
 * Receives a B-MAC packet.
 * 
 * @param src [out] source address
 * @param dst [out] destination address
 * @param data [out] payload of the packet
 * @param data_length [in] maximal size of the payload
 *                    [out] actual size of the payload
 * @param ms [in] maximal duration (in ms) to wait for a packet
 * @return 0 on success, -1 on timeout.
 */
int bmac_receive(u_short *src, u_short *dst, u_char *data, 
                 u_short *data_length, u_long ms);

/**
 * Registers some B-MAC related btnode terminal-commands.
 */
void bmac_cmds_register_cmds(void);

/**
 * MAC-Interface
 */
extern mac_interface_t bmac_interface;

#endif /* _BMAC_H_ */
